/* 
 ==============================================
     Steuerung Controllino - Loxone UDP
 ==============================================
Folgende Zeile sind zu bearbeiten: 43-52, 69, 83, 84 (Entprellzeit der Eingänge, möglicherweise anzupassen)

Analoge Ausgänge bei Loxone:
AO1=<v.1>   AO2=<v.1>
Relay Ausgang:  Block A(oben)   Block B(unten)
                R_A=<v.1>       R_B=<v.1>
Digital Ausgang:
DO=<v.1>
Digitale Eingänge:  DI0-3    AI0-AI7      AI8-AI11    (AI0-AI11 ungetestet)
                    10\v     AI_D1=\v    AI_D2=\v
Analoge Eingänge:
AI0=\v    AI1=\v    [...] (ungetestet)

Watchdog (Minutenimpuls):
online


*/
#include <Controllino.h>  /* Usage of CONTROLLINO library allows you to use CONTROLLINO_xx aliases in your sketch. */
//#include <SPI.h>
#include <Ethernet.h>
#include <EthernetUdp.h>
//#include <Wire.h>
#include <string.h> 


//#define DEBUG Serial    // Herauskommentieren, wenn keine serielle Ausgabe erwünscht wird
#ifdef DEBUG
    #define debug(...) DEBUG.print(__VA_ARGS__)
    #define debugln(...) DEBUG.println(__VA_ARGS__)
    #define debugbegin(...) DEBUG.begin(__VA_ARGS__)
  #else
    #define debug(...)
    #define debugln(...)
    #define debugbegin(...)
#endif

// Netzwerkeinstellungen: --------------------------------------------------------------------------------------Network Settings--------------------------------------------------------------------------------------
  byte mac[] = { 0xDB, 0xAD, 0xBE, 0xAF, 0xAE, 0xAD };
  IPAddress ip(192, 168, 2, 3);               // IP-Adresse eingeben
  IPAddress gateway(192, 168, 2, 1);          // Gateway eingeben
  IPAddress subnet(255, 255, 255, 0);         // Subnetz eingeben
// Local UDP port to listen on
  unsigned int localPort = 7010;              // Empfangs UDP Port festlegen
// Recipient IP
  IPAddress RecipientIP(192, 168, 178, 2);    // Loxone IP
// Recipient UDP Port
  unsigned int RecipientPort = 7011;          // Sende UDP Port festlegen
// EthernetUDP instance to send and receive packets over UDP
  EthernetUDP Udp;
// Watchdog -------------------------------------------------------------------
  unsigned long wd_t;
  unsigned long wd_pre = 60000;
// buffers for receiving and sending data ------------------------------------
  char packetBuffer[UDP_TX_PACKET_MAX_SIZE];
  char  ReplyBuffer[] = "acknowledged";       // a string to send back
// Variablen: ----------------------------------------------------------------
  char input[] = "";
  char value_v[] = "";
  char delimiter[] = "=";
  int ao1_value, ao1_lastvalue = 0;
  int ao2_value, ao2_lastvalue = 0;
  int do_value = 0;
  unsigned long di0_millis,di1_millis,di2_millis,di3_millis,act_millis;
  unsigned long bouncetime = 100;
  bool do0_value,do1_value,do2_value,do3_value,do4_value,do5_value,do6_value,do7_value=LOW;
  bool do0_lastvalue,do1_lastvalue,do2_lastvalue,do3_lastvalue,do4_lastvalue,do5_lastvalue,do6_lastvalue,do7_lastvalue;
  bool di0_value, di1_value, di2_value, di3_value = LOW;
  int di_udp,di_udp_last,di_udp_send;
  int r_A_value, r_B_value = 0;
  bool r0_value,r1_value,r2_value,r3_value,r4_value,r5_value,r6_value,r7_value,r8_value,r9_value=LOW;
  bool r0_lastvalue,r1_lastvalue,r2_lastvalue,r3_lastvalue,r4_lastvalue,r5_lastvalue,r6_lastvalue,r7_lastvalue,r8_lastvalue,r9_lastvalue;
  int ai0_value,ai1_value,ai2_value,ai3_value,ai4_value,ai5_value,ai6_value,ai7_value,ai8_value,ai9_value,ai10_value,ai11_value;
  byte aiD1_value, aiD2_value = 0;
  byte aiD1_lastvalue, aiD2_lastvalue;
  
    
    //Festlegen der Eingangsart für AI0-AI11 (Digital oder Analog)
    char ai2_art, ai3_art, ai4_art, ai5_art, ai6_art, ai7_art, ai8_art, ai9_art, ai10_art, ai11_art = "Digital";
    char ai0_art, ai1_art = "Analog";

    

void setup() {
// For debug only
  Serial.begin(9600);
// Ein-/Ausgänge: -------------------------------------------------------------
  pinMode(CONTROLLINO_AO0, OUTPUT);
  pinMode(CONTROLLINO_AO1, OUTPUT);
  pinMode(CONTROLLINO_DO0, OUTPUT);
  pinMode(CONTROLLINO_DO1, OUTPUT);
  pinMode(CONTROLLINO_DO2, OUTPUT);
  pinMode(CONTROLLINO_DO3, OUTPUT);
  pinMode(CONTROLLINO_DO4, OUTPUT);
  pinMode(CONTROLLINO_DO5, OUTPUT);
  pinMode(CONTROLLINO_DO6, OUTPUT);
  pinMode(CONTROLLINO_DO7, OUTPUT);

  pinMode(CONTROLLINO_R0, OUTPUT);
  pinMode(CONTROLLINO_R1, OUTPUT);
  pinMode(CONTROLLINO_R2, OUTPUT);
  pinMode(CONTROLLINO_R3, OUTPUT);
  pinMode(CONTROLLINO_R4, OUTPUT);
  pinMode(CONTROLLINO_R5, OUTPUT);
  pinMode(CONTROLLINO_R6, OUTPUT);
  pinMode(CONTROLLINO_R7, OUTPUT);
  pinMode(CONTROLLINO_R8, OUTPUT);
  pinMode(CONTROLLINO_R9, OUTPUT);
  
  pinMode(CONTROLLINO_DI0, INPUT);
  pinMode(CONTROLLINO_DI1, INPUT);
  pinMode(CONTROLLINO_DI2, INPUT);
  pinMode(CONTROLLINO_DI3, INPUT);
  
  pinMode(CONTROLLINO_AI0, INPUT);
  pinMode(CONTROLLINO_AI1, INPUT);
  pinMode(CONTROLLINO_AI2, INPUT);
  pinMode(CONTROLLINO_AI3, INPUT);
  pinMode(CONTROLLINO_AI4, INPUT);
  pinMode(CONTROLLINO_AI5, INPUT);
  pinMode(CONTROLLINO_AI6, INPUT);
  pinMode(CONTROLLINO_AI7, INPUT);
  pinMode(CONTROLLINO_AI8, INPUT);
  pinMode(CONTROLLINO_AI9, INPUT);
  pinMode(CONTROLLINO_AI10, INPUT);
  pinMode(CONTROLLINO_AI11, INPUT);
  
// Start Ethernet
  Ethernet.begin(mac, ip);
// Start UDP
  Udp.begin(localPort);
//Send UDP Ready
  sendUDP("UDP Ready");

debugln("Setup abgeschlossen"); 
}

void loop() {
//Empfange:
  int packetSize = Udp.parsePacket();
    if (packetSize) {
      // read the packet into packetBufffer
      Udp.read(packetBuffer, UDP_TX_PACKET_MAX_SIZE);
      debugln(packetBuffer);
    
      char *ptr;
      ptr = strtok(packetBuffer, delimiter);
      strcpy(input, ptr);
    
//AO1================================================================================================================= 
          if (strcmp(input, "AO1") == 0) {
                    ptr = strtok(NULL, delimiter);       //Pointer auf Wert im char
                    strcpy(value_v, ptr);                //Kopiere Wert in Variable value_v
                    ao1_value = atoi(value_v);           //Wandle value_v in Integer
                    debug("Empfange AO1 = "); debugln(ao1_value);
          }
//AO2=================================================================================================================  
          else if (strcmp(input, "AO2") == 0) {
                    ptr = strtok(NULL, delimiter);    //Pointer auf Wert im char
                    strcpy(value_v, ptr);             //Kopiere Wert in Variable value_v
                    ao2_value = atoi(value_v);            //Wandle value_v in Integer
                    debug("Empfange AO2 = "); debugln(ao2_value);
          }
/*  ======================================================================================================================
 *  Für Erweiterung über PWM Ausgänge: (1 Ausgangsbeispiel)
            else if (strcmp(input, "AO5") == 0) {
                 ptr = strtok(NULL, delimiter);     //Pointer auf Wert im char
                 strcpy(value_v, ptr);              //Kopiere Wert in Variable value_v
                 ao5_value = atoi(value_v);             //Wandle value_v in Integer
            analogWrite(AO5, value);                //Schreibe PWM Ausgang
            debug("SET AO5 = "), debugln(value);
          }
DO=====================================================================================================================================*/  
          else if (strcmp(input, "DO") == 0) {
                    ptr = strtok(NULL, delimiter);    //Pointer auf Wert im char
                    strcpy(value_v, ptr);             //Kopiere Wert in Variable value_v
                    do_value = atoi(value_v);            //Wandle value_v in Integer
                    debug("Empfange DO = "); debugln(do_value);
         }
         else if (strcmp(input, "R_A") == 0) {
                    ptr = strtok(NULL, delimiter);    //Pointer auf Wert im char
                    strcpy(value_v, ptr);             //Kopiere Wert in Variable value_v
                    r_A_value = atoi(value_v);            //Wandle value_v in Integer
                    debug("Empfange R_A = "); debugln(r_A_value);
         }
         else if (strcmp(input, "R_B") == 0) {
                    ptr = strtok(NULL, delimiter);    //Pointer auf Wert im char
                    strcpy(value_v, ptr);             //Kopiere Wert in Variable value_v
                    r_B_value = atoi(value_v);            //Wandle value_v in Integer
                    debug("Empfange R_B = "); debugln(r_B_value);
         }
      else {
    debugln("Kein Input gefunden");
  }

  char input[] = "";
  char value_v[] = "";
  }
//Verarbeite Empfangene Daten:  
  if(ao1_value < 0){ ao1_value = 0; }
  else if(ao1_value > 1000) { ao1_value = 1000; }
  if(ao2_value < 0){ ao1_value = 0; }
  else if(ao2_value > 1000) { ao2_value = 1000; }

  if(ao1_value != ao1_lastvalue){
    analogWrite(CONTROLLINO_AO0, map(ao1_value,0,1000,0,255));
    ao1_lastvalue = ao1_value;
  }
  if(ao2_value != ao2_lastvalue){
    analogWrite(CONTROLLINO_AO1, map(ao2_value,0,1000,0,255));
    ao2_lastvalue = ao2_value; 
  }

do0_value = bitRead(do_value, 0);
do1_value = bitRead(do_value, 1);
do2_value = bitRead(do_value, 2);
do3_value = bitRead(do_value, 3);
do4_value = bitRead(do_value, 4);
do5_value = bitRead(do_value, 5);
do6_value = bitRead(do_value, 6);
do7_value = bitRead(do_value, 7);

//Relay Output Block B
r0_value = bitRead(r_B_value, 0);
r1_value = bitRead(r_B_value, 1);
r2_value = bitRead(r_B_value, 2);
r3_value = bitRead(r_B_value, 3);
r4_value = bitRead(r_B_value, 4);
r5_value = bitRead(r_B_value, 5);
//Relay Output Block A
r6_value = bitRead(r_A_value, 0);
r7_value = bitRead(r_A_value, 1);
r8_value = bitRead(r_A_value, 2);
r9_value = bitRead(r_A_value, 3);


if(do0_value != do0_lastvalue) { digitalWrite(CONTROLLINO_DO0, do0_value); do0_lastvalue=do0_value; }
if(do1_value != do1_lastvalue) { digitalWrite(CONTROLLINO_DO1, do1_value); do1_lastvalue=do1_value; }
if(do2_value != do2_lastvalue) { digitalWrite(CONTROLLINO_DO2, do2_value); do2_lastvalue=do2_value; }
if(do3_value != do3_lastvalue) { digitalWrite(CONTROLLINO_DO3, do3_value); do3_lastvalue=do3_value; }
if(do4_value != do4_lastvalue) { digitalWrite(CONTROLLINO_DO4, do4_value); do4_lastvalue=do4_value; }
if(do5_value != do5_lastvalue) { digitalWrite(CONTROLLINO_DO5, do5_value); do5_lastvalue=do5_value; }
if(do6_value != do6_lastvalue) { digitalWrite(CONTROLLINO_DO6, do6_value); do6_lastvalue=do6_value; }
if(do7_value != do7_lastvalue) { digitalWrite(CONTROLLINO_DO7, do7_value); do7_lastvalue=do7_value; }

if(r0_value != r0_lastvalue) { digitalWrite(CONTROLLINO_R0, r0_value); r0_lastvalue=r0_value; }
if(r1_value != r1_lastvalue) { digitalWrite(CONTROLLINO_R1, r1_value); r1_lastvalue=r1_value; }
if(r2_value != r2_lastvalue) { digitalWrite(CONTROLLINO_R2, r2_value); r2_lastvalue=r2_value; }
if(r3_value != r3_lastvalue) { digitalWrite(CONTROLLINO_R3, r3_value); r3_lastvalue=r3_value; }
if(r4_value != r4_lastvalue) { digitalWrite(CONTROLLINO_R4, r4_value); r4_lastvalue=r4_value; }
if(r5_value != r5_lastvalue) { digitalWrite(CONTROLLINO_R5, r5_value); r5_lastvalue=r5_value; }
if(r6_value != r6_lastvalue) { digitalWrite(CONTROLLINO_R6, r6_value); r6_lastvalue=r6_value; }
if(r7_value != r7_lastvalue) { digitalWrite(CONTROLLINO_R7, r7_value); r7_lastvalue=r7_value; }
if(r8_value != r8_lastvalue) { digitalWrite(CONTROLLINO_R8, r8_value); r8_lastvalue=r8_value; }
if(r9_value != r9_lastvalue) { digitalWrite(CONTROLLINO_R9, r9_value); r9_lastvalue=r9_value; }


act_millis = millis();

if ((act_millis-di0_millis) >= bouncetime){
        di0_value = digitalRead(CONTROLLINO_DI0);
        debug("In1: ");debugln(di0_value);
            bitWrite(di_udp,0,di0_value);
            di0_millis = millis();
        }
if ((act_millis-di1_millis) >= bouncetime){
        di1_value = digitalRead(CONTROLLINO_DI1);
            bitWrite(di_udp,1,di1_value);
            di1_millis = millis();
        }
if ((act_millis-di2_millis) >= bouncetime){
        di2_value = digitalRead(CONTROLLINO_DI2);
            bitWrite(di_udp,2,di2_value);
            di2_millis = millis();
        }
if ((act_millis-di3_millis) >= bouncetime){
        di3_value = digitalRead(CONTROLLINO_DI3);
            bitWrite(di_udp,3,di3_value);
            di3_millis = millis();
        }

debug("di_udp: "); debug(di_udp); debug("   di_udp_last=: "); debugln(di_udp_last);
if(di_udp != di_udp_last){
        debug("Sende DI = "); debugln(di_udp);
        di_udp_send = di_udp + 10000;    // Als Kanal 10 senden (Loxone Eingangsbefehl: 10\v)
        Udp.beginPacket(RecipientIP, RecipientPort); //Sende an UDP Empfänger
        Udp.print(di_udp_send);
        Udp.endPacket();
        di_udp_last= di_udp;
}

if(ai0_art == "Analog"){ ai0_value = analogRead(CONTROLLINO_AI0); sendUDP("AI0=" + ai0_value);}
else { bitWrite(aiD1_value, 0, digitalRead(CONTROLLINO_AI0)); }
if(ai1_art == "Analog"){ ai1_value = analogRead(CONTROLLINO_AI1); sendUDP("AI1=" + ai1_value);}
else { bitWrite(aiD1_value, 1, digitalRead(CONTROLLINO_AI1)); }
if(ai2_art == "Analog"){ ai2_value = analogRead(CONTROLLINO_AI2); sendUDP("AI2=" + ai2_value);}
else { bitWrite(aiD1_value, 2, digitalRead(CONTROLLINO_AI2)); }
if(ai3_art == "Analog"){ ai3_value = analogRead(CONTROLLINO_AI3); sendUDP("AI3=" + ai3_value);}
else { bitWrite(aiD1_value, 3, digitalRead(CONTROLLINO_AI3)); }
if(ai4_art == "Analog"){ ai4_value = analogRead(CONTROLLINO_AI4); sendUDP("AI4=" + ai4_value);}
else { bitWrite(aiD1_value, 4, digitalRead(CONTROLLINO_AI4)); }
if(ai5_art == "Analog"){ ai5_value = analogRead(CONTROLLINO_AI5); sendUDP("AI5=" + ai5_value);}
else { bitWrite(aiD1_value, 5, digitalRead(CONTROLLINO_AI5)); }
if(ai6_art == "Analog"){ ai6_value = analogRead(CONTROLLINO_AI6); sendUDP("AI6=" + ai6_value);}
else { bitWrite(aiD1_value, 6, digitalRead(CONTROLLINO_AI6)); }
if(ai7_art == "Analog"){ ai7_value = analogRead(CONTROLLINO_AI7); sendUDP("AI7=" + ai7_value);}
else { bitWrite(aiD1_value, 7, digitalRead(CONTROLLINO_AI7)); }
if(ai8_art == "Analog"){ ai8_value = analogRead(CONTROLLINO_AI8); sendUDP("AI8=" + ai8_value);}
else { bitWrite(aiD2_value,0, digitalRead(CONTROLLINO_AI8)); }
if(ai9_art == "Analog"){ ai9_value = analogRead(CONTROLLINO_AI9); sendUDP("AI9=" + ai9_value);}
else { bitWrite(aiD2_value,1, digitalRead(CONTROLLINO_AI9)); }
if(ai10_art == "Analog"){ ai10_value = analogRead(CONTROLLINO_AI10); sendUDP("AI10=" + ai10_value);}
else { bitWrite(aiD2_value,2, digitalRead(CONTROLLINO_AI10)); }
if(ai11_art == "Analog"){ ai11_value = analogRead(CONTROLLINO_AI11); sendUDP("AI11=" + ai11_value);}
else { bitWrite(aiD2_value,3, digitalRead(CONTROLLINO_AI11)); }

if(aiD1_value != aiD1_lastvalue){
  sendUDP("AI_D1=" + aiD1_value);
  aiD1_lastvalue = aiD1_value;
}
if(aiD2_value != aiD2_lastvalue){
  sendUDP("AI_D2=" + aiD2_value);
  aiD2_lastvalue = aiD2_value;  
}

  //Watchdog:
  if (act_millis - wd_t >= wd_pre) {
    wd_t = millis();
    sendUDP("online");
  }

//Loop Ende
}

// Function to send UDP packets
void sendUDP(String text)
{
  Udp.beginPacket(RecipientIP, RecipientPort);
  // Udp.write("Test");
  Udp.print(text);
  Udp.endPacket();
}
